class BBOX:
    def __init__(self, bbox):
        self.left = bbox[0]
        self.top = bbox[1]
        self.right = bbox[2]
        self.bottom = bbox[3]
        self.width = self.right - self.left
        self.height = self.bottom - self.top
        self.center_x = int(self.width * 0.5)
        self.center_y = int(self.height * 0.5)

    def crop(self, img):
        return img[self.top:self.bottom, self.left:self.right]

    def __call__(self, *args, **kwargs):
        return self.left, self.top, self.right, self.bottom



try:
    import cv2
    import skimage.exposure
    import numpy as np
except ImportError:
    raise 'pip install ninjatools[image] or ninjatools[all] to use image functions!'


class ImageProcessor:

    def __init__(self):
        self.TRACKBAR_WINDOW = "Trackbars"
        self.GUI_enabled = False

    def init_control_gui(self):
        cv2.namedWindow(self.TRACKBAR_WINDOW, cv2.WINDOW_GUI_NORMAL)
        cv2.setWindowProperty(self.TRACKBAR_WINDOW, cv2.WND_PROP_TOPMOST, 1)
        cv2.setWindowProperty(self.TRACKBAR_WINDOW, cv2.WINDOW_AUTOSIZE, 1)

        cv2.namedWindow('Smoothing Trackbars', cv2.WINDOW_GUI_NORMAL)
        cv2.setWindowProperty('Smoothing Trackbars', cv2.WND_PROP_TOPMOST, 1)
        cv2.setWindowProperty('Smoothing Trackbars', cv2.WINDOW_AUTOSIZE, 1)

        # Required callback. we'll be using getTrackbarPos() to do lookups
        # Instead of using the callback.
        def nothing(_):
            pass

        # Create trackbars for RGB filter
        cv2.createTrackbar('RMin', self.TRACKBAR_WINDOW, 0, 255, nothing)
        cv2.createTrackbar('GMin', self.TRACKBAR_WINDOW, 0, 255, nothing)
        cv2.createTrackbar('BMin', self.TRACKBAR_WINDOW, 0, 255, nothing)
        cv2.createTrackbar('RMax', self.TRACKBAR_WINDOW, 0, 255, nothing)
        cv2.createTrackbar('GMax', self.TRACKBAR_WINDOW, 0, 255, nothing)
        cv2.createTrackbar('BMax', self.TRACKBAR_WINDOW, 0, 255, nothing)

        # Set default value for RGB filter
        cv2.setTrackbarPos('RMax', self.TRACKBAR_WINDOW, 255)
        cv2.setTrackbarPos('GMax', self.TRACKBAR_WINDOW, 255)
        cv2.setTrackbarPos('BMax', self.TRACKBAR_WINDOW, 255)

        # Create trackbars for bracketing.
        cv2.createTrackbar('HMin', self.TRACKBAR_WINDOW, 0, 179, nothing)
        cv2.createTrackbar('SMin', self.TRACKBAR_WINDOW, 0, 255, nothing)
        cv2.createTrackbar('VMin', self.TRACKBAR_WINDOW, 0, 255, nothing)
        cv2.createTrackbar('HMax', self.TRACKBAR_WINDOW, 0, 179, nothing)
        cv2.createTrackbar('SMax', self.TRACKBAR_WINDOW, 0, 255, nothing)
        cv2.createTrackbar('VMax', self.TRACKBAR_WINDOW, 0, 255, nothing)

        # Set default value for Max HSV trackbars
        cv2.setTrackbarPos('HMax', self.TRACKBAR_WINDOW, 179)
        cv2.setTrackbarPos('SMax', self.TRACKBAR_WINDOW, 255)
        cv2.setTrackbarPos('VMax', self.TRACKBAR_WINDOW, 255)

        # Trackbars for increasing/decreasing saturation and value
        cv2.createTrackbar('SAdd', self.TRACKBAR_WINDOW, 0, 255, nothing)
        cv2.createTrackbar('SSub', self.TRACKBAR_WINDOW, 0, 255, nothing)
        cv2.createTrackbar('VAdd', self.TRACKBAR_WINDOW, 0, 255, nothing)
        cv2.createTrackbar('VSub', self.TRACKBAR_WINDOW, 0, 255, nothing)

        # Trackbars for edge creation
        cv2.createTrackbar('KernelSize', self.TRACKBAR_WINDOW, 0, 30, nothing)
        cv2.createTrackbar('ErodeIter', self.TRACKBAR_WINDOW, 0, 5, nothing)
        cv2.createTrackbar('DilateIter', self.TRACKBAR_WINDOW, 0, 5, nothing)
        cv2.createTrackbar('Canny1', self.TRACKBAR_WINDOW, 0, 200, nothing)
        cv2.createTrackbar('Canny2', self.TRACKBAR_WINDOW, 0, 500, nothing)

        # Trackbars for Smoothness
        cv2.createTrackbar('blackAndWhite', 'Smoothing Trackbars', 0, 1, nothing)
        cv2.createTrackbar('b_wX', 'Smoothing Trackbars', 125, 255, nothing)
        cv2.createTrackbar('b_wY', 'Smoothing Trackbars', 255, 255, nothing)
        cv2.createTrackbar('smoothing', 'Smoothing Trackbars', 0, 1, nothing)
        cv2.createTrackbar('sigmaX', 'Smoothing Trackbars', 3, 100, nothing)
        cv2.createTrackbar('sigmaY', 'Smoothing Trackbars', 3, 100, nothing)
        cv2.createTrackbar('in_rangeX', 'Smoothing Trackbars', 125, 255, nothing)
        cv2.createTrackbar('in_rangeY', 'Smoothing Trackbars', 255, 255, nothing)
        cv2.createTrackbar('out_rangeX', 'Smoothing Trackbars', 0, 255, nothing)
        cv2.createTrackbar('out_rangeY', 'Smoothing Trackbars', 255, 255, nothing)

    # Returns an RGB filter object based on the control GUI values
    def get_rgb_filter_from_controls(self):
        # Get current positions of all trackbars
        rgb_values = FilterValues()
        rgb_values.r_min = cv2.getTrackbarPos('RMin', self.TRACKBAR_WINDOW)
        rgb_values.g_min = cv2.getTrackbarPos('GMin', self.TRACKBAR_WINDOW)
        rgb_values.b_min = cv2.getTrackbarPos('BMin', self.TRACKBAR_WINDOW)
        rgb_values.r_max = cv2.getTrackbarPos('RMax', self.TRACKBAR_WINDOW)
        rgb_values.g_max = cv2.getTrackbarPos('GMax', self.TRACKBAR_WINDOW)
        rgb_values.b_max = cv2.getTrackbarPos('BMax', self.TRACKBAR_WINDOW)
        return rgb_values

    # Returns an HSV filter object based on the control GUI values
    def get_hsv_filter_from_controls(self):
        # Get current positions of all trackbars
        hsv_values = FilterValues()
        hsv_values.hMin = cv2.getTrackbarPos('HMin', self.TRACKBAR_WINDOW)
        hsv_values.sMin = cv2.getTrackbarPos('SMin', self.TRACKBAR_WINDOW)
        hsv_values.vMin = cv2.getTrackbarPos('VMin', self.TRACKBAR_WINDOW)
        hsv_values.hMax = cv2.getTrackbarPos('HMax', self.TRACKBAR_WINDOW)
        hsv_values.sMax = cv2.getTrackbarPos('SMax', self.TRACKBAR_WINDOW)
        hsv_values.vMax = cv2.getTrackbarPos('VMax', self.TRACKBAR_WINDOW)
        hsv_values.sAdd = cv2.getTrackbarPos('SAdd', self.TRACKBAR_WINDOW)
        hsv_values.sSub = cv2.getTrackbarPos('SSub', self.TRACKBAR_WINDOW)
        hsv_values.vAdd = cv2.getTrackbarPos('VAdd', self.TRACKBAR_WINDOW)
        hsv_values.vSub = cv2.getTrackbarPos('VSub', self.TRACKBAR_WINDOW)
        return hsv_values

    # Returns a Canny edge filter object based on the control GUI values
    def get_edge_filter_from_controls(self):
        # Get current positions of all trackbars
        edge_filter = FilterValues()
        edge_filter.kernelSize = cv2.getTrackbarPos('KernelSize', self.TRACKBAR_WINDOW)
        edge_filter.erodeIter = cv2.getTrackbarPos('ErodeIter', self.TRACKBAR_WINDOW)
        edge_filter.dilateIter = cv2.getTrackbarPos('DilateIter', self.TRACKBAR_WINDOW)
        edge_filter.canny1 = cv2.getTrackbarPos('Canny1', self.TRACKBAR_WINDOW)
        edge_filter.canny2 = cv2.getTrackbarPos('Canny2', self.TRACKBAR_WINDOW)
        return edge_filter

    # Returns a Smoothing filter object based on the control GUI values
    @staticmethod
    def get_smoothing_filter_from_controls():
        # Get current positions of all trackbars
        smoothing_filter = FilterValues()
        smoothing_filter.blackAndWhite = cv2.getTrackbarPos('blackAndWhite', 'Smoothing Trackbars')
        smoothing_filter.b_wX = cv2.getTrackbarPos('b_wX', 'Smoothing Trackbars')
        smoothing_filter.b_wY = cv2.getTrackbarPos('b_wY', 'Smoothing Trackbars')
        smoothing_filter.smoothing = cv2.getTrackbarPos('smoothing', 'Smoothing Trackbars')
        smoothing_filter.sigmaX = cv2.getTrackbarPos('sigmaX', 'Smoothing Trackbars')
        smoothing_filter.sigmaY = cv2.getTrackbarPos('sigmaY', 'Smoothing Trackbars')
        smoothing_filter.in_rangeX = cv2.getTrackbarPos('in_rangeX', 'Smoothing Trackbars')
        smoothing_filter.in_rangeY = cv2.getTrackbarPos('in_rangeY', 'Smoothing Trackbars')
        smoothing_filter.out_rangeX = cv2.getTrackbarPos('out_rangeX', 'Smoothing Trackbars')
        smoothing_filter.out_rangeY = cv2.getTrackbarPos('out_rangeY', 'Smoothing Trackbars')

        return smoothing_filter

    # apply adjustments to an HSV channel
    # https://stackoverflow.com/questions/49697363/shifting-hsv-pixel-values-in-python-using-numpy
    @staticmethod
    def shift_channel(c, amount):
        if amount > 0:
            lim = 255 - amount
            c[c >= lim] = 255
            c[c < lim] += amount
        elif amount < 0:
            amount = -amount
            lim = amount
            c[c <= lim] = 0
            c[c > lim] -= amount
        return c

    def apply_filter(self, image, filter_values=None, hsv_first: bool = True):

        if hsv_first:
            image = self.apply_hsv_filter(image, filter_values)
            image = self.apply_rgb_filter(image, filter_values)
        else:
            image = self.apply_rgb_filter(image, filter_values)
            image = self.apply_hsv_filter(image, filter_values)
        image = self.apply_edge_filter(image, filter_values)
        image = self.apply_smoothing_filter(image, filter_values)

        return image

    def apply_rgb_filter(self, image, rgb_mask_filter=None):
        # if we haven't been given a defined filter, use the filter values from the GUI
        if not rgb_mask_filter:
            if not self.GUI_enabled:
                self.GUI_enabled = True
                self.init_control_gui()
            rgb_mask_filter = self.get_rgb_filter_from_controls()

        # extract bgr channels
        bgr = image[:, :, 0:3]

        # hsv = cv2.cvtColor(image, cv2.COLOR_BGR2HSV)

        # select purple
        rgb_min = np.array([rgb_mask_filter.r_min, rgb_mask_filter.g_min, rgb_mask_filter.b_min], dtype=np.uint8)
        rgb_max = np.array([rgb_mask_filter.r_max, rgb_mask_filter.g_max, rgb_mask_filter.b_max], dtype=np.uint8)

        # second mask
        rgb_min2 = np.array([255, 255, 0], dtype=np.uint8)
        rgb_max2 = np.array([255, 255, 0], dtype=np.uint8)

        # Apply the thresholds
        mask = cv2.inRange(bgr, rgb_min, rgb_max)
        mask2 = cv2.inRange(bgr, rgb_min2, rgb_max2)

        mask = cv2.bitwise_or(mask2, mask)
        result = cv2.bitwise_and(bgr, bgr, mask=mask)

        return result

    def apply_hsv_filter(self, original_image, hsv_filter=None):
        # convert image to HSV
        hsv = cv2.cvtColor(original_image, cv2.COLOR_BGR2HSV)

        # if we haven't been given a defined filter, use the filter values from the GUI
        if not hsv_filter:
            if not self.GUI_enabled:
                self.GUI_enabled = True
                self.init_control_gui()
            hsv_filter = self.get_hsv_filter_from_controls()

        # add/subtract saturation and value
        h, s, v = cv2.split(hsv)
        s = self.shift_channel(s, hsv_filter.sAdd)
        s = self.shift_channel(s, -hsv_filter.sSub)
        v = self.shift_channel(v, hsv_filter.vAdd)
        v = self.shift_channel(v, -hsv_filter.vSub)
        hsv = cv2.merge([h, s, v])

        # Set minimum and maximum HSV values to display
        lower = np.array([hsv_filter.hMin, hsv_filter.sMin, hsv_filter.vMin])
        upper = np.array([hsv_filter.hMax, hsv_filter.sMax, hsv_filter.vMax])

        bgr = original_image[:, :, 0:3]

        # second mask
        rgb_min2 = np.array([255, 255, 0], dtype=np.uint8)
        rgb_max2 = np.array([255, 255, 0], dtype=np.uint8)

        # Apply the thresholds
        mask = cv2.inRange(hsv, lower, upper)
        mask2 = cv2.inRange(bgr, rgb_min2, rgb_max2)
        mask = cv2.bitwise_or(mask, mask2)

        result = cv2.bitwise_and(hsv, hsv, mask=mask)

        # convert back to BGR for imshow() to display it properly
        img = cv2.cvtColor(result, cv2.COLOR_HSV2BGR)

        return img

    def apply_edge_filter(self, original_image, edge_filter=None):
        # if we haven't been given a defined filter, use the filter values from the GUI
        if not edge_filter:
            edge_filter = self.get_edge_filter_from_controls()

        kernel = np.ones((edge_filter.kernelSize, edge_filter.kernelSize), np.uint8)
        image = cv2.erode(original_image, kernel, iterations=edge_filter.erodeIter)
        image = cv2.dilate(image, kernel, iterations=edge_filter.dilateIter)

        if edge_filter.canny1 > 0 or edge_filter.canny2 > 0:
            # canny edge detection
            image = cv2.Canny(image, edge_filter.canny1, edge_filter.canny2)

            # convert single channel image back to BGR
            image = cv2.cvtColor(image, cv2.COLOR_GRAY2BGR)

        return image

    def apply_smoothing_filter(self, original_image, smoothing_filter=None):
        if not smoothing_filter:
            smoothing_filter = self.get_smoothing_filter_from_controls()

        if smoothing_filter.blackAndWhite == 1:
            r1, g1, b1 = 0, 0, 0  # Value to be retained
            r2, g2, b2 = 255, 255, 255  # Value to be replaced

            red, green, blue = original_image[:, :, 0], original_image[:, :, 1], original_image[:, :, 2]
            mask = (red != r1) & (green != g1) & (blue != b1)
            original_image[:, :, :3][mask] = [r2, g2, b2]

        if smoothing_filter.smoothing == 1:
            sigma_x = smoothing_filter.sigmaX if smoothing_filter.sigmaX > 0 else 1
            sigma_y = smoothing_filter.sigmaY if smoothing_filter.sigmaX > 0 else 1
            original_image = cv2.GaussianBlur(original_image, (0, 0),
                                              sigmaX=sigma_x, sigmaY=sigma_y,
                                              borderType=cv2.BORDER_DEFAULT
                                              )

            original_image = skimage.exposure.rescale_intensity(original_image,
                                                                in_range=(
                                                                    smoothing_filter.in_rangeX,
                                                                    smoothing_filter.in_rangeY
                                                                ),
                                                                out_range=(
                                                                    smoothing_filter.out_rangeX,
                                                                    smoothing_filter.out_rangeY)
                                                                )

            original_image = cv2.convertScaleAbs(original_image)

        return original_image


class FilterValues:
    def __init__(self, argv=(0, 0, 0, 255, 255, 255, 0, 0,
                             0, 179, 255, 255, 0, 0, 0, 0,
                             5, 1, 1, 100, 200, 0, 127, 255, 0, 3, 3, 125, 255, 0, 255)):
        self.r_min, self.g_min, self.b_min, self.r_max, self.g_max, self.b_max, self.hMin, self.sMin, self.vMin, \
        self.hMax, self.sMax, self.vMax, self.sAdd, self.sSub, self.vAdd, self.vSub, self.kernelSize, \
        self.erodeIter, self.dilateIter, self.canny1, self.canny2, self.blackAndWhite, self.b_wX, \
        self.b_wY, self.smoothing, self.sigmaX, self.sigmaY, self.in_rangeX, self.in_rangeY, self.out_rangeX, \
        self.out_rangeY = argv

    def __call__(self, *args, **kwargs):
        return self.r_min, self.g_min, self.b_min, self.r_max, self.g_max, self.b_max, self.hMin, self.sMin, \
               self.vMin, \
               self.hMax, self.sMax, self.vMax, self.sAdd, self.sSub, self.vAdd, self.vSub, self.kernelSize, \
               self.erodeIter, self.dilateIter, self.canny1, self.canny2, self.blackAndWhite, self.b_wX, \
               self.b_wY, self.smoothing, self.sigmaX, self.sigmaY, self.in_rangeX, self.in_rangeY, self.out_rangeX, \
               self.out_rangeY



from typing import List

try:
    import psutil
    import pymem
except ImportError:
    raise 'pip install ninjatools[memory] or ninjatools[all]  to use image functions!'

from ninja_tools.utils import Utilities

Utilities.try_import('ReadWriteMemory', 'git+https://github.com/vsantiago113/ReadWriteMemory')

from ReadWriteMemory import ReadWriteMemory


class Parser:
    def __init__(self, pid: int):
        self.pid = pid

        self.rwm = ReadWriteMemory()
        self.process = self.rwm.get_process_by_id(pid)
        self.process.open()

        self.process_name = self.process.name

        self.pm = pymem.Pymem()
        self.pm.open_process_from_id(pid)
        self.base_address = pymem.process.module_from_name(
            self.pm.process_handle, self.process_name).lpBaseOfDll

    # Readers
    def read(self, address):
        # return self.process.read(address)
        return self.pm.read_int(address)

    def read_str(self, address: int, length: int):
        # return self.process.readString(address, length)
        return self.pm.read_string(address, length)

    def read_byte(self, address: int, length: int = 1):
        # return self.process.readByte(address, length)
        return self.pm.read_bytes(address, length)

    # Writers
    def write(self, address: int, integer: int):
        # return self.process.write(address, integer)
        return self.pm.write_int(address, integer)

    def write_str(self, address: int, string: str):
        # return self.process.writeString(address, string)
        return self.pm.write_string(address, string)

    def write_byte(self, address: int, bytes_: List[hex]):
        return self.process.writeByte(address, bytes_)
        # return self.pm.write_bytes(address, bytes_)

    # Resolvers
    def resolve(self, address: int, offsets: List):
        return self.process.get_pointer(address, offsets=offsets)

    def resolver(self, address: int, offsets: List):
        # return self.process.read(self.resolve(address, offsets))
        return self.pm.read_int(self.resolve(address, offsets))

    def resolver_str(self, address: int, length: int, offsets: List):
        # return self.read_str(self.resolve(address, offsets), length)
        return self.pm.read_string(self.resolve(address, offsets), length)

    def resolver_byte(self, address: int, length: int = 1, offsets: List = ()):
        # return self.read_byte(self.resolve(address, offsets), length)
        return self.pm.read_bytes(self.resolve(address, offsets), length)



import warnings

try:
    import openpyxl
    import webcolors
    from openpyxl.utils.cell import coordinate_from_string, column_index_from_string
except ImportError:
    raise 'pip install ninjatools[excel] or ninjatools[all] to use excel functions!'

# Remove warnings
warnings.filterwarnings('ignore', category=UserWarning, module='openpyxl')


class ColorObject(object):
    def __init__(self, hex_, rgb, color_name):
        self.hex = hex_
        self.rgb = rgb
        self.name = color_name

    def __repr__(self):
        str_color = ', '.join(str(_) for _ in self.rgb)
        return f'Closest Color: {self.name} | RGB: ({str_color})'


class Excel:
    def __init__(self, workbook_path):
        self.wb = openpyxl.load_workbook(workbook_path, data_only=True)
        self.sheet_name = None

    def get_sheets(self) -> list:
        """
        Returns a list of sheet names
        :return:
        """
        return [_ for _ in self.wb.sheetnames]

    # Get cell_value using openpyxl
    def cell(self, cell, sheet_name=None):
        """
        Returns the value of a cell
        :param cell:
        :param sheet_name:
        :return:
        """

        if isinstance(cell, str):
            cell = self.get_cell(cell)
        elif not isinstance(cell, tuple) and not isinstance(cell, list):
            raise TypeError('cell must be a tuple or string')

        sheet_name = sheet_name if sheet_name else self.sheet_name
        ws = self.wb[sheet_name]
        value = ws.cell(cell[0], cell[1]).value
        value = '' if value is None else value
        value = str(value).strip().replace('\xa0', ' ')
        return value

    def get_color(self, cell) -> ColorObject:
        """
        Returns the color of a cell
        :param cell:
        :return:
        """
        ws = self.wb[self.sheet_name]
        hex_color = ws[cell].fill.start_color.index
        hex_color = f'#{hex_color[2:]}'
        rgb_color = webcolors.hex_to_rgb(hex_color)
        rgb_color = (rgb_color.red, rgb_color.green, rgb_color.blue)
        color_name = self.closest_color(rgb_color)

        return ColorObject(hex_color, rgb_color, color_name)

    def read_range(self, start, end, sheet_name=None, remove_empty=False) -> list:
        """
        Reads a range of cells
        :param start:
        :param end:
        :param sheet_name:
        :param remove_empty:
        :return:
        """
        data = []

        sheet_name = sheet_name if sheet_name else self.sheet_name

        rc_start = self.get_cell(start)
        rc_end = self.get_cell(end)

        for row in range(rc_start[0], rc_end[0] + 1):
            temp = []
            for col in range(rc_start[1], rc_end[1] + 1):
                temp.append(self.cell((row, col), sheet_name))

            if remove_empty:
                if not all([not _ for _ in temp]):
                    data.append(temp)
            else:
                data.append(temp)

        return data

    # Still buggy
    def read_indefinitely(self, start_cell: str, num_of_columns: int, num_of_rows=None, steps: int = 1) -> list:
        """
        Reads a range of cells indefinitely
        :param start_cell:
        :param num_of_columns:
        :param num_of_rows:
        :param steps:
        :return:
        """

        data = []

        ord_num = ord(start_cell[0])
        i = int(start_cell[1:])

        while True:
            try:
                data.append([self.cell(f'{chr(ord_num + _)}{i}') for _ in range(num_of_columns)])
                if num_of_rows:
                    if len(data) >= num_of_rows:
                        break
            except (IndexError, AttributeError):
                if not data:
                    print(f'Check if the cell {start_cell} to # of columns is not all empty!')
                break
            except (Exception,):
                break
            i += steps

        return data

    def get_last(self) -> str:
        """
        Returns the last cells
        :return:
        """
        ws = self.wb[self.sheet_name]
        return self.get_cell_name(ws.max_row, ws.max_column)

    @staticmethod
    def get_cell(cell) -> tuple:
        """
        Returns the cell coordinates
        :param cell:
        :return:
        """
        xy = coordinate_from_string(cell)
        col = column_index_from_string(xy[0])
        row = xy[1]
        return row, col

    @staticmethod
    def get_cell_name(row, column) -> str:
        """
        Returns the cell name
        :param row:
        :param column:
        :return:
        """
        return openpyxl.utils.get_column_letter(column) + str(row)

    @staticmethod
    def convert_to_dict(data: list, key_idx: int, header_idx: int = 0) -> dict:
        """
        Converts a list to a dictionary
        :param data:
        :param key_idx:
        :param header_idx:
        :return:
        """
        dict_data = {}
        for idx, row in enumerate(data):
            if idx > 0:
                dict_data[row[key_idx]] = dict(zip(data[header_idx], row))

        return dict_data

    @staticmethod
    def closest_color(rgb):
        differences = {}
        for color_hex, color_name in webcolors.CSS3_HEX_TO_NAMES.items():
            r, g, b = webcolors.hex_to_rgb(color_hex)
            differences[sum([(r - rgb[0]) ** 2,
                             (g - rgb[1]) ** 2,
                             (b - rgb[2]) ** 2])] = color_name
        return differences[min(differences.keys())]

    # TODO: Cell write values/formulas



import ctypes
import subprocess
import sys
from datetime import datetime
from math import sqrt
from time import perf_counter, sleep

import pkg_resources
import psutil
import pyperclip
import win32gui
import win32process

from ninja_tools.bbox import BBOX

GetWindowTextLength = ctypes.windll.user32.GetWindowTextLengthW
GetWindowText = ctypes.windll.user32.GetWindowTextW


class Utilities:
    ##########################
    # Import functions
    ##########################
    @staticmethod
    def try_import(package, installer=None):
        def check(_):
            return _ in [_.project_name for _ in pkg_resources.working_set]

        if not check(package):
            installer = package if not installer else installer
            install = subprocess.Popen([sys.executable, "-m", "pip", "install", installer])
            install.wait()

    ##########################
    # Clipboard Functions
    ##########################
    @staticmethod
    def put_on_clipboard(_):
        pyperclip.copy(_)

    @staticmethod
    def get_from_clipboard():
        return pyperclip.paste()

    ##########################
    # Math Functions
    ##########################
    @staticmethod
    def safe_div(x, y):
        return 0 if y == 0 else x / y

    def safe_div_int(self, x, y):
        return int(self.safe_div(x, y))

    def safe_div_round(self, x, y, decimals=2):
        return round(self.safe_div(x, y), decimals)

    @staticmethod
    def get_distance(p0, p1):
        return sqrt((p0[0] - p1[0]) ** 2 + (p0[1] - p1[1]) ** 2)

    def get_distance_int(self, p0, p1):
        return int(self.get_distance(p0, p1))

    def get_distance_rounded(self, p0, p1, decimals=0):
        return round(self.get_distance(p0, p1), decimals)

    ##########################
    # Process Functions
    ##########################
    @staticmethod
    def get_handle_from_title(window):
        return win32gui.FindWindow(None, window)

    @staticmethod
    def get_handle_from_pid(pid):
        def callback(hwnd, _):
            __, found_pid = win32process.GetWindowThreadProcessId(hwnd)

            if found_pid == pid:
                _.append(hwnd)
            return True

        handle = []
        win32gui.EnumWindows(callback, handle)
        return handle[0]

    @staticmethod
    def get_process_name_from_pid(pid):
        return psutil.Process(pid).name()

    def get_pid_from_title(self, window):
        handle = self.get_handle_from_title(window)
        _, found_pid = win32process.GetWindowThreadProcessId(handle)
        return found_pid

    @staticmethod
    def get_current_window_title():
        return win32gui.GetWindowText(win32gui.GetForegroundWindow())

    def get_window_title_from_pid(self, pid):
        handle = self.get_handle_from_pid(pid)
        return self.get_window_title_from_handle(handle)

    def is_current_window_title(self, window_name: str):
        return window_name == self.get_current_window_title()

    @staticmethod
    def get_window_title_from_handle(hwnd):
        length = GetWindowTextLength(hwnd)
        buff = ctypes.create_unicode_buffer(length + 1)
        GetWindowText(hwnd, buff, length + 1)
        return buff.value

    @staticmethod
    def _get_rect(type_, handle):
        if type_ == "WINDOW":
            return win32gui.GetWindowRect(handle)
        elif type_ == "CLIENT":
            return win32gui.GetClientRect(handle)
        else:
            raise 'Either type == WINDOW or CLIENT only!'

    def get_window_rect_from_handle(self, handle, bbox: bool = True):
        rect = self._get_rect("WINDOW", handle)
        return BBOX(rect) if bbox else rect

    def get_window_rect_from_pid(self, pid, bbox: bool = True):
        handle = self.get_handle_from_pid(pid)
        rect = self._get_rect("WINDOW", handle)
        return BBOX(rect) if bbox else rect

    def get_client_rect_from_handle(self, handle, bbox: bool = True):
        rect = self._get_rect("CLIENT", handle)
        return BBOX(rect) if bbox else rect

    def get_client_rect_from_pid(self, pid, bbox: bool = True):
        handle = self.get_handle_from_pid(pid)
        rect = self._get_rect("CLIENT", handle)
        return BBOX(rect) if bbox else rect

    ##########################
    # Utility Functions
    ##########################
    @staticmethod
    def find(input_, string, options=any, contains=True):
        if isinstance(input_, list):
            if contains:
                return options(item in string for item in input_)
            else:
                return options(item not in string for item in input_)
        else:
            if contains:
                return string in input_
            else:
                return input_ not in string

    @staticmethod
    def pause(milliseconds: int):
        sleep(milliseconds * 0.001)

    @staticmethod
    def make_hash(d):
        __ = ''
        for _ in d:
            __ += str(d[_])
        return hash(__)

    @staticmethod
    def timestamp():
        (dt, micro) = datetime.utcnow().strftime('%Y%m%d-%H%M%S.%f').split('.')
        dt = "%s.%03d" % (dt, int(micro) * 0.001)
        return dt

    @staticmethod
    def datetime():
        return datetime.utcnow()

    @staticmethod
    def perf():
        return perf_counter()

    ##########################
    # I/O
    ##########################
    @staticmethod
    def write_to_file(filename, text, method: str = "a", add_new_line: bool = True):
        with open(filename, method) as file:
            if add_new_line:
                file.write(text + "\n")
            else:
                file.write(text)

    @staticmethod
    def read_file(filename, method="r"):
        return open(filename, method).read()

    @staticmethod
    def read_lines(filename, method="r"):
        return open(filename, method, encoding="utf8").readlines()

    ##########################
    # Assorted
    ##########################
    @staticmethod
    def cv2_show(cv2, image, window_name=None, delay=1, stop_key='q'):
        cv2.imshow(window_name, image)
        if cv2.waitKey(delay) & 0xFF == ord(stop_key):
            cv2.destroyAllWindows()

    # def timeout(self, key, ms):
    #     if key in self.timeouts:
    #         return self.timeouts[key].timeout(ms)
    #
    #     else:
    #         self.timeouts[key] = ProcessTime()
    #         return False



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="NinjaTools",
    version="0.8.15",
    author="Nikko Gonzales",
    author_email="nikkoxgonzales@gmail.com",
    description="Bunch of useful tools",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/nikkoxgonzales/ninja-tools",
    install_requires=[
        'pywin32>=303',
        'pyperclip>=1.8.2',
        'psutil>=5.9.0',
    ],
    extras_require={
        'image': [
            'opencv-python==4.5.5.62',
            'scikit-image>=0.19.2',
            'numpy>=1.22.3',
        ],

        'memory': [
            'Pymem>=1.8.5',
            'regex'
        ],

        'excel': [
            'openpyxl==3.0.10',
            'webcolors>=1.12'
        ],

        'all': [
            'opencv-python==4.5.5.62',
            'scikit-image>=0.19.2',
            'openpyxl==3.0.10',
            'webcolors>=1.12'
            'numpy>=1.22.3',
            'Pymem>=1.8.5',
        ]
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: Microsoft :: Windows",
    ],
    packages=setuptools.find_packages(),
    python_requires='>=3.9'
)



try:
    import numpy as np
except ImportError:
    raise 'pip install ninjatools[image] or ninjatools[all] to use image functions!'


class ImageFunctions:
    @staticmethod
    def most_rgb(img):
        colors, count = np.unique(img.reshape(-1, img.shape[-1]), axis=0, return_counts=True)
        b, g, r = colors[count.argmax()]
        return b, g, r

    def transparency(self, img):
        r, g, b = self.most_rgb(img)
        img[np.all(img == (r, g, b), axis=-1)] = (255, 255, 255)
        h, w, c = img.shape

        # append Alpha channel -- required for BGRA (Blue, Green, Red, Alpha)
        bgra = np.concatenate([img, np.full((h, w, 1), 255, dtype=np.uint8)], axis=-1)

        # create a mask where white pixels ([255, 255, 255]) are True
        white = np.all(img == [255, 255, 255], axis=-1)

        # change the values of Alpha to 0 for all the white pixels
        bgra[white, -1] = 0

        return bgra

    @staticmethod
    def h_split(img):
        h, w = img.shape[:2]

        h_cut = h // 2
        upper = img[:h_cut, :]
        lower = img[h_cut:, :]

        return upper, lower

    @staticmethod
    def w_split(img):
        h, w = img.shape[:2]

        w_cut = w // 2
        left = img[:w_cut, :]
        right = img[w_cut:, :]

        return left, right
